/***************************************************************************
 *
 * gpm.c : GNOME Power Manager
 *
 * Copyright (C) 2005 Richard Hughes, <richard@hughsie.com>
 *
 * Taken in part from:
 * - lshal   (C) 2003 David Zeuthen, <david@fubar.dk>
 * - notibat (C) 2004 Benjamin Kahn, <xkahn@zoned.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 **************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>
#include <dbus/dbus-glib.h>

#include <X11/X.h>

#include <gdk/gdkx.h>
#include <gnome.h>
#include <glade/glade.h>
#include <gconf/gconf-client.h>

#include <libhal.h>
#include "common.h"
#include "clipboard.h"
#include "icons.h"
#include "gpm.h"
#include "gpn.h"

/* static */
static LibHalContext *hal_ctx;
static gboolean useSystemBus, isVerbose;

/* no need for IPC with globals */
HasData has_data;
StateData state_data;

GPtrArray *objectData;

/** Do an interactive alert
 *
 *  @param  text		the text to be used in the dialogue
 * 				e.g. "Logitech MX-1000 mouse"
 */
static void
do_interactive_alert (const gchar *text)
{
	g_assert (text);
	g_debug ("do_interactive_alert");

	GtkWidget *widget = gnome_message_box_new (text, 
                                GNOME_MESSAGE_BOX_WARNING,
                                GNOME_STOCK_BUTTON_OK, 
                                NULL);
	gtk_window_set_title (GTK_WINDOW (widget), "Power Manager Alert!");
	gtk_widget_show (widget);
}

/** Gets policy from gconf
 *
 *  @param  name		gconf policy name
 *  @return 			the int gconf value of the policy
 */
gint
get_policy_string (const gchar *name)
{
	g_assert (name);
	GConfClient *client = gconf_client_get_default ();
	gchar *gconfpath = g_strconcat ("/apps/gnome-power/policy/", name, NULL);
	gchar *valuestr = gconf_client_get_string (client, gconfpath, NULL);
	gint value = convert_string_to_policy (valuestr);
	g_free (gconfpath);
	g_free (valuestr);
	return value;
}

#if FUTURE_CODE
/** Gets policy from gconf
 *
 *  @param  name		gconf policy name
 *  @return 			the int gconf value of the policy
 */
static int
get_policy_int (const gchar *name)
{
	g_assert (name);
	GConfClient *client = gconf_client_get_default ();
	gchar *gconfpath = g_strconcat ("/apps/gnome-power/policy/", name, NULL);
	gint value = gconf_client_get_int (client, gconfpath, NULL);
	g_free (gconfpath);
	return value;
}
#endif

static void
run_gconf_script (const char *path)
{
	GConfClient *client = gconf_client_get_default ();
	gchar *command = gconf_client_get_string (client, path, NULL);
	if (command) {
		g_debug ("Executing '%s'", command);
#if GPM_DO_ACTIONS
		if (!g_spawn_command_line_async (command, NULL))
			g_warning ("Couldn't execute '%s'.\n", command);
#endif
		g_free (command);
	} else
		g_warning ("'%s' is missing!", path);
}

/** Do the action dictated by policy from gconf
 *
 *  @param  policy_number	What to do!
 */
void
action_policy_do (gint policy_number)
{
	if (policy_number == ACTION_NOTHING) {
		g_debug ("*ACTION+DBUS* Doing nothing");
	} else if (policy_number == ACTION_WARNING) {
		g_debug ("*ACTION+DBUS* Send warning");
		do_interactive_alert ("*Testing*");
	} else if (policy_number == ACTION_REBOOT) {
		g_debug ("*ACTION+DBUS* Reboot");
		run_gconf_script ("/apps/gnome-power/scripts/reboot");
	} else if (policy_number == ACTION_SUSPEND) {
		g_debug ("*ACTION+DBUS* Suspend");
		run_gconf_script ("/apps/gnome-power/scripts/suspend");
	} else if (policy_number == ACTION_HIBERNATE) {
		g_debug ("*ACTION+DBUS* Hibernate");
		run_gconf_script ("/apps/gnome-power/scripts/hibernate");
	} else if (policy_number == ACTION_SHUTDOWN) {
		g_debug ("*ACTION+DBUS* Shutdown");
		run_gconf_script ("/apps/gnome-power/scripts/shutdown");
	} else if (policy_number == ACTION_BATTERY_CHARGE) {
		g_debug ("*ACTION+DBUS* Battery Charging");
	} else if (policy_number == ACTION_BATTERY_DISCHARGE) {
		g_debug ("*ACTION+DBUS* Battery Discharging");
	} else if (policy_number == ACTION_NOW_BATTERYPOWERED) {
		g_debug ("*DBUS* Now battery powered");
	} else if (policy_number == ACTION_NOW_MAINSPOWERED) {
		g_debug ("*DBUS* Now mains powered");
	} else
		g_warning ("action_policy_do called with unknown action %i", 
								policy_number);
}

/** Compare the old and the new values, if different or force'd then updates gconf
 *
 *  @param  name		general gconf name, e.g. hasButtons
 */
static void
compare_bool_set_gconf (const gchar *name, gboolean *has_dataold, gboolean *has_datanew, gboolean force)
{
	g_assert (name);
	gchar *gconfpath = g_strconcat ("/apps/gnome-power/general/", name, NULL);
	GConfClient *client = gconf_client_get_default ();

	if (force || *has_datanew != *has_dataold) {
		*has_dataold = *has_datanew;
		gconf_client_set_bool (client, gconfpath, *has_datanew, NULL);
		g_debug ("%s = %i", name, *has_datanew);
	}
	g_free (gconfpath);
}

/** Description
 *
 *  @param  xxxx		test
 */
#if 0
static void
do_policy_percent (int action)
{
	if (action == ACTION_POWER_STATE_CHANGE) {
		g_debug ("DBUS: %s = %i", "onBatteryPower", state_data.onBatteryPower);
		if (has_data.hasDisplays && has_data.hasLCD) {
			if (state_data.onBatteryPower)
				action_policy_percent ("LCDBatteryPercentage");
			else
				action_policy_percent ("LCDMainsPercentage");
		}
		if (has_data.hasBatteries) {
			if (state_data.onBatteryPower)
				action_policy_do ("BatteryFailAction");
		}
	}
}
#endif

/** Recalculate logic of StateData, without any DBUS, all cached internally
 * Exported DBUS interface values goes here :-)
 *  @param  coldplug		If set, send events even if they are the same
 */
static void
update_state_logic (GPtrArray *parray, gboolean coldplug)
{
	g_assert (parray);
	g_debug ("update_state_logic");
	gint a;
	GenericObject *slotData;
	/* set up temp. state */
	StateData state_datanew;
	state_datanew.onBatteryPower = FALSE;
	state_datanew.onUPSPower = FALSE;

	for (a=0;a<parray->len;a++) {
		slotData = (GenericObject *) g_ptr_array_index (parray, a);
		if (slotData->powerDevice == POWER_UPS && slotData->isDischarging)
			state_datanew.onUPSPower = TRUE;
		if (slotData->powerDevice == POWER_PRIMARY_BATTERY && slotData->isDischarging)
			state_datanew.onBatteryPower = TRUE;
	}

	/* get old value */
	if (has_data.hasBatteries == TRUE) {
		/* Reverse logic as only one ac_adapter is needed to be "on mains power" */
		for (a=0;a<parray->len;a++) {
			slotData = (GenericObject *) g_ptr_array_index (parray, a);
/** 
 ** TODO: valgrind says Conditional jump or move depends on uninitialised value(s) on 222 (->present?)
 **/
			if (slotData->powerDevice == POWER_AC_ADAPTER && slotData->present) {
				state_datanew.onBatteryPower = FALSE;
				break;
				}
		}
	} else {
		g_debug ("Cannot be on batteries if have none...");
		state_datanew.onBatteryPower = FALSE;
	}

	if (coldplug || state_datanew.onBatteryPower != state_data.onBatteryPower) {
		state_data.onBatteryPower = state_datanew.onBatteryPower;
		/*do_policy_percent (ACTION_POWER_STATE_CHANGE);*/
		if (state_data.onBatteryPower)
			action_policy_do (ACTION_NOW_BATTERYPOWERED);
		else
			action_policy_do (ACTION_NOW_MAINSPOWERED);
#if 0
		if (has_data.hasDisplays) {
			if (state_data.onBatteryPower)
				action_policy_percent ("lcd_batteries");
			else
				action_policy_percent ("lcd_mains");
		}
		if (has_data.hasBatteries) {
			if (state_data.onBatteryPower)
				action_policy_do ("combo_battery_fail");
		}
#endif
	}

	if (coldplug || state_datanew.onUPSPower != state_data.onUPSPower) {
		state_data.onUPSPower = state_datanew.onUPSPower;
		g_debug ("DBUS: %s = %i", "onUPSPower", state_data.onUPSPower);
	}

}

/** Recalculate logic of HasData, without any DBUS, all cached internally
 *  All complicated convoluted logic goes here :-)
 *  @param  coldplug		If set, send events even if they are the same
 */
static void
update_has_logic (GPtrArray *parray, gboolean coldplug)
{
	g_assert (parray);
	g_debug ("update_has_logic");
		
	gint a;
	GenericObject *slotData;
	/* set up temp. state */
	HasData has_datanew;
	has_datanew.hasBatteries = FALSE;
	has_datanew.hasAcAdapter = FALSE;
	has_datanew.hasButtonPower = TRUE;
	has_datanew.hasButtonSleep = TRUE;
	has_datanew.hasButtonLid = TRUE;
	has_datanew.hasUPS = FALSE;
	has_datanew.hasDisplays = TRUE;
	has_datanew.hasHardDrive = TRUE;
	has_datanew.hasLCD = FALSE;

	for (a=0;a<parray->len;a++) {
		slotData = (GenericObject *) g_ptr_array_index (parray, a);
		if (slotData->powerDevice == POWER_UPS)
			has_datanew.hasUPS = TRUE;
		else if (slotData->powerDevice == POWER_MOUSE)
			has_datanew.hasMouse = TRUE;
		else if (slotData->powerDevice == POWER_KEYBOARD)
			has_datanew.hasKeyboard = TRUE;
		else if (slotData->powerDevice == POWER_PRIMARY_BATTERY)
			if (slotData->present)
				has_datanew.hasBatteries = TRUE;
			else
				g_debug ("Battery missing?!?");
		else if (slotData->powerDevice == POWER_AC_ADAPTER)
			has_datanew.hasAcAdapter = TRUE;
	}
	compare_bool_set_gconf ("hasUPS", 
		&has_data.hasUPS, &has_datanew.hasUPS, coldplug);
	compare_bool_set_gconf ("hasButtonPower", 
		&has_data.hasButtonPower, &has_datanew.hasButtonPower, coldplug);
	compare_bool_set_gconf ("hasButtonSleep", 
		&has_data.hasButtonSleep, &has_datanew.hasButtonSleep, coldplug);
	compare_bool_set_gconf ("hasButtonLid", 
		&has_data.hasButtonLid, &has_datanew.hasButtonLid, coldplug);
	compare_bool_set_gconf ("hasAcAdapter", 
		&has_data.hasAcAdapter, &has_datanew.hasAcAdapter, coldplug);
	compare_bool_set_gconf ("hasBatteries", 
		&has_data.hasBatteries, &has_datanew.hasBatteries, coldplug);
	compare_bool_set_gconf ("hasLCD", 
		&has_data.hasLCD, &has_datanew.hasLCD, coldplug);
	compare_bool_set_gconf ("hasHardDrive", 
		&has_data.hasHardDrive, &has_datanew.hasHardDrive, coldplug);
	compare_bool_set_gconf ("hasDisplays", 
		&has_data.hasDisplays, &has_datanew.hasDisplays, coldplug);
}

/** Convenience function.
 *  Prints errors due to wrong values expected (exposes bugs, rather than hides them)
 */
static void
dbus_error_print (DBusError *error)
{
	g_assert (error);
	if (dbus_error_is_set (error)) {
		g_warning ("DBUS:%s", error->message);
		dbus_error_free (error);
	}
}

/** Generic exit
 *
 */
static void
gpm_exit (void)
{
	g_debug ("Quitting!");
	gpn_icon_destroy (ICON_ALL);
	exit (0);
}

/** Prints the objectData table with any parameters
 *
 *  @param  parray		pointer array to GenericObject
 */
void
genericobject_print (GPtrArray *parray)
{
	g_assert (parray);
	int a;
	GenericObject *slotData;
	for (a=0;a<parray->len;a++) {
		slotData = (GenericObject *) g_ptr_array_index (parray, a);
		g_assert (slotData);
		g_print ("[%i] udi: %s\n", a, slotData->udi);
		g_print ("     powerDevice: %i\n", slotData->powerDevice);
	}
	
}

/** Finds a device from the objectData table
 *
 *  @param  parray		pointer array to GenericObject
 *  @param  udi			HAL UDI
 */
static gint
find_udi_parray_index (GPtrArray *parray, const char *udi)
{
	GenericObject *slotData;
	int a;
	for (a=0;a<parray->len;a++) {
		slotData = (GenericObject *) g_ptr_array_index (parray, a);
		g_assert (slotData);
		if (strcmp (slotData->udi, udi) == 0)
			return a;
	}
	return -1;
}

/** Finds a device from the objectData table
 *
 *  @param  parray		pointer array to GenericObject
 *  @param  udi			HAL UDI
 */
static GenericObject *
genericobject_find (GPtrArray *parray, const char *udi)
{
	int a = find_udi_parray_index (parray, udi);
	if (a != -1)
		return (GenericObject *) g_ptr_array_index (parray, a);
	return NULL;
}

/** Adds a device to the objectData table *IF DOES NOT EXIST*
 *
 *  @param  parray		pointer array to GenericObject
 *  @param  udi			HAL UDI
 *  @return			TRUE if we added to the table
 */
static GenericObject *
genericobject_add (GPtrArray *parray, const char *udi)
{
	int a = find_udi_parray_index (parray, udi);
	if (a != -1)
		return NULL;

	GenericObject *slotData = g_new (GenericObject, 1);
	strcpy (slotData->udi, udi);
	slotData->powerDevice = POWER_UNKNOWN;
	slotData->slot = a;
	g_ptr_array_add (parray, (gpointer) slotData);
	return slotData;
}

/** Adds a ac_adapter device. Also sets up properties on cached object
 *
 *  @param  udi			UDI
 */
static void
add_ac_adapter (const gchar *udi)
{
	DBusError error;
	g_assert (udi);
	GenericObject *slotData = genericobject_add (objectData, udi);
	if (slotData) {
		slotData->powerDevice = POWER_AC_ADAPTER;
		slotData->percentageCharge = 0;
		g_debug ("Device '%s' added", udi);
		/* ac_adapter batteries might be missing */
		dbus_error_init (&error);
		slotData->present = libhal_device_get_property_bool (hal_ctx,
			udi, "ac_adapter.present", &error);
		dbus_error_print (&error);
	} else
		g_warning ("Device '%s' already added!", udi);
}

/** Adds a battery device, of any type. Also sets up properties on cached object
 *
 *  @param  udi			UDI
 */
static void
add_battery (const gchar *udi)
{
	g_assert (udi);
	gint powerDevice;
	DBusError error;
	gchar *type = NULL;

	GenericObject *slotData = genericobject_add (objectData, udi);
	if (!slotData) {
		g_warning ("Cannot add object to table!");
		return;
	}

	dbus_error_init (&error);
	type = libhal_device_get_property_string (hal_ctx, udi, "battery.type", &error);
	dbus_error_print (&error);
	if (!type)
		return;
	/* battery is further refined using the .type property */
	powerDevice = convert_haltype_to_powerdevice (type);
	libhal_free_string (type);

	GString *device = convert_powerdevice_to_string (powerDevice);
	if (device) {
		g_debug ("%s added", device->str);
		g_string_free (device, TRUE);
		device = NULL;
	}

	/* set cached variables up */
	dbus_error_init (&error);
	slotData->minutesRemaining = libhal_device_get_property_int (hal_ctx,
		udi, "battery.remaining_time", &error) / 60;
	dbus_error_print (&error);

	/*
	 * We need the RAW readings so we keep functions modular and 
	 * acpi/apm neutral
	 */
	dbus_error_init (&error);
	slotData->rawCharge = libhal_device_get_property_int (hal_ctx,
		udi, "battery.charge_level.current", &error);
	dbus_error_print (&error);

	/* 
	 * We need the RAW readings so we can process them later on
	 * for time remaining
	 */
	dbus_error_init (&error);
	slotData->rawLastFull = libhal_device_get_property_int (hal_ctx,
		udi, "battery.charge_level.last_full", &error);
	dbus_error_print (&error);

	/* PMU/ACPI batteries might be missing */
	dbus_error_init (&error);
	slotData->present = libhal_device_get_property_bool (hal_ctx,
		udi, "battery.present", &error);
	dbus_error_print (&error);

	/* battery might not be rechargeable, have to check */
	dbus_error_init (&error);
	slotData->isRechargeable = libhal_device_get_property_bool (hal_ctx,
		udi, "battery.is_rechargeable", &error);
	dbus_error_print (&error);
	if (slotData->isRechargeable) {
		dbus_error_init (&error);
		slotData->isCharging = libhal_device_get_property_bool (hal_ctx,
			udi, "battery.rechargeable.is_charging", &error);
		dbus_error_print (&error);
		dbus_error_init (&error);
		slotData->isDischarging = libhal_device_get_property_bool (hal_ctx,
			udi, "battery.rechargeable.is_discharging", &error);
		dbus_error_print (&error);
	}

	slotData->powerDevice = powerDevice;
	update_percentage_charge (slotData);
}

/** Coldplugs devices of type battery & ups at startup
 *
 */
static void
coldplug_devices (void)
{
	gint i, num_devices;
	char **device_names;
	DBusError error;

	/* devices of type battery */
	dbus_error_init (&error);
	device_names = libhal_find_device_by_capability (hal_ctx, "battery", 
							&num_devices, &error);
	dbus_error_print (&error);
	if (device_names == NULL)
		g_warning ("Couldn't obtain list of batteries");
	for (i = 0; i < num_devices; i++)
		add_battery (device_names[i]);
	libhal_free_string_array (device_names);

	/* devices of type ac_adapter */
	dbus_error_init (&error);
	device_names = libhal_find_device_by_capability (hal_ctx, "ac_adapter",
							&num_devices, &error);
	dbus_error_print (&error);
	if (device_names == NULL)
		g_warning ("Couldn't obtain list of ac_adapters");
	for (i = 0; i < num_devices; i++)
		add_ac_adapter (device_names[i]);
	libhal_free_string_array (device_names);
}

/** Removes any type of device
 *
 *  @param  udi			UDI
 */
static void
remove_devices (const char *udi)
{
	g_assert (udi);
	int a = find_udi_parray_index (objectData, udi);
	if (a == -1) {
		g_warning ("Asked to remove '%s' when not present", udi);
		return;
	}
	g_debug ("Removed '%s'", udi);
	g_ptr_array_remove_index (objectData, a);
}

/** Invoked when a device is removed from the Global Device List. Simply
 *  prints a message on stderr.
 *
 *  @param  udi			UDI
 */
static void
device_removed (LibHalContext *ctx, const char *udi)
{
	g_assert (udi);
	g_debug ("device_removed");
	/*
	 * UPS's/mice/keyboards don't use battery.present
	 * they just disappear from the device tree
	 */
	remove_devices (udi);
	/* our state has changed, update */
	update_has_logic (objectData, FALSE);
	update_state_logic (objectData, FALSE);
	if (isVerbose)
		genericobject_print (objectData);
	gpn_icon_update (ICON_ALL);
}

/** Invoked when device in the Global Device List acquires a new capability.
 *  Prints the name of the capability to stderr.
 *
 *  @param  udi			UDI
 *  @param  capability          Name of capability
 */
static void
device_new_capability (LibHalContext *ctx, const char *udi, const char *capability)
{
	g_assert (udi);
	g_assert (capability);
	g_debug ("device_new_capability");
	/*
	 * UPS's/mice/keyboards don't use battery.present
	 * they just appear in the device tree
	 */
	if (strcmp (capability, "battery") == 0) {
		add_battery (udi);
		/* our state has changed, update */
		update_has_logic (objectData, FALSE);
		update_state_logic (objectData, FALSE);
		gpn_icon_update (ICON_ALL);
	}
}

/** Invoked when a property of a device in the Global Device List is
 *  changed, and we have we have subscribed to changes for that device.
 *
 *  @param  udi                 Univerisal Device Id
 *  @param  key                 Key of property
 */
static void
property_modified (LibHalContext *ctx, const char *udi, const char *key,
		   dbus_bool_t is_removed, dbus_bool_t is_added)
{
	g_assert (udi);
	g_assert (key);
	g_debug ("property_modified");
	DBusError error;
	GenericObject *slotData;

	/* only process modified entries, not added or removed keys */
	if (is_removed||is_added)
		return;

	/* no point continuing any further if we are never going to match ...*/
	if (strncmp (key, "battery", 7) != 0 && strncmp (key, "ac_adapter", 10) != 0)
		return;

	slotData = genericobject_find (objectData, udi);
	
#if 0
	dbus_error_init (&error);
	gchar *type = libhal_device_get_property_string (hal_ctx, udi, "info.category", &error);
	dbus_error_print (&error);
	libhal_free_string (type);
		g_warning ("Cannot find UDI '%s' in the objectData", udi);
		if (powerDevice == POWER_AC_ADAPTER)) {
			add_ac_adapter (udi);
			object_table_find (udi, &slotData, objectData);
		} else if (powerDevice != POWER_BATTERY || strcmp (type, "hiddev") == 0)  {
			add_battery (udi);
			object_table_find (udi, &slotData, objectData);
		} else if (strcmp (type, "button") == 0)  {
			/* we should do nothing here - we do not cache the button state */
			return;
		} else {
			g_warning ("Unrecognised category '%s'!", type);
			libhal_free_string (type);
			return;
		}
		libhal_free_string (type);
	}
#endif

	/* if we BUG here then *HAL* has a problem where key modification is
	 * done before capability is present
	 */
	g_assert (slotData);
	gboolean updateHas = FALSE;
	gboolean updateState = FALSE;
	
	g_debug ("key = '%s'", key);
	g_debug ("udi = '%s'", udi);
	if (strcmp (key, "battery.present") == 0 || strcmp (key, "ac_adapter.present") == 0) {
		dbus_error_init (&error);
		slotData->present = libhal_device_get_property_bool (ctx, udi, key, &error);
		dbus_error_print (&error);
		updateHas = TRUE;
		updateState = TRUE;
	} else if (strcmp (key, "battery.rechargeable.is_charging") == 0) {
		dbus_error_init (&error);
		slotData->isCharging = libhal_device_get_property_bool (ctx, udi, key, &error);
		dbus_error_print (&error);
		updateState = TRUE;
	} else if (strcmp (key, "battery.rechargeable.is_discharging") == 0) {
		dbus_error_init (&error);
		slotData->isDischarging = libhal_device_get_property_bool (ctx, udi, key, &error);
		dbus_error_print (&error);
		updateState = TRUE;
	} else if (strcmp (key, "battery.charge_level.current") == 0) {
		dbus_error_init (&error);
		slotData->rawCharge = libhal_device_get_property_int (ctx, udi, key, &error);
		dbus_error_print (&error);
	} else if (strcmp (key, "battery.remaining_time") == 0) {
		dbus_error_init (&error);
		slotData->minutesRemaining = libhal_device_get_property_int (ctx, udi, key, &error) / 60;
		dbus_error_print (&error);
	} else if (strcmp (key, "battery.charge_level.rate") == 0) {
		/* ignore */
	} else {
		g_warning ("Cannot recognise key '%s' from UDI '%s'", key, udi);
		return;
	}

	if (updateHas)
		update_has_logic (objectData, TRUE);
	if (updateState)
		update_state_logic (objectData, TRUE);

	update_percentage_charge (slotData);
	gpn_icon_update (ICON_ALL);
}

/** Invoked when a property of a device in the Global Device List is
 *  changed, and we have we have subscribed to changes for that device.
 *
 *  @param  udi                 Univerisal Device Id
 *  @param  condition_name      Name of condition
 *  @param  message             D-BUS message with parameters
 */
static void
device_condition (LibHalContext *ctx,
		  const char *udi, 
		  const char *condition_name,
		  const char *condition_details)
{
	g_assert (udi);
	DBusError error;
	gchar *type;

	if (strcmp (condition_name, "ButtonPressed") == 0) {
		dbus_error_init (&error);
		type = libhal_device_get_property_string (hal_ctx, udi, "button.type", &error);
		dbus_error_print (&error);
		g_debug ("ButtonPressed : %s", type);
		if (strcmp (type, "power") == 0) {
			action_policy_do (get_policy_string ("ButtonPower"));
		} else if (strcmp (type, "sleep") == 0) {
			action_policy_do (get_policy_string ("ButtonSuspend"));
		} else if (strcmp (type, "lid") == 0) {
			gboolean value;
			/* we only do a lid event when the lid is OPENED */
			dbus_error_init (&error);
			value = libhal_device_get_property_bool (ctx, udi, "button.state.value", &error);
			dbus_error_print (&error);
			if (value)
				action_policy_do (get_policy_string ("ButtonLid"));
		} else
			g_warning ("Button '%s' unrecognised", type);
		libhal_free_string (type);
	}
}

/** LibHal setup
 *
 *  @param  conn		D-BUS connection
 *  @return			D-BUS connection or NULL on error
 */
static DBusConnection *
libhal_setup ()
{
	g_debug ("libhal_setup");

	DBusConnection *conn = NULL;
	DBusError error;
	dbus_error_init (&error);
	conn = dbus_bus_get (DBUS_BUS_SYSTEM, &error);
	if (conn == NULL) {
		g_warning ("error: dbus_bus_get: %s: %s\n", error.name, error.message);
		return NULL;
	}
	if ((hal_ctx = libhal_ctx_new ()) == NULL) {
		g_warning ("error: libhal_ctx_new\n");
		return NULL;
	}
	if (!libhal_ctx_set_dbus_connection (hal_ctx, conn)) {
		g_warning ("error: libhal_ctx_set_dbus_connection: %s: %s\n", error.name, error.message);
		return NULL;
	}
	if (!libhal_ctx_init (hal_ctx, &error)) {
		g_warning ("error: libhal_ctx_init: %s: %s\n", error.name, error.message);
		return NULL;
	}
	libhal_ctx_set_device_property_modified (hal_ctx, property_modified);
	libhal_ctx_set_device_removed (hal_ctx, device_removed);
	libhal_ctx_set_device_new_capability (hal_ctx, device_new_capability);
	libhal_ctx_set_device_condition (hal_ctx, device_condition);
	return conn;
}

/** LibHal cleanup
 *
 *  @param  conn		D-BUS connection
 */
static void
libhal_cleanup (DBusConnection *conn)
{
	g_assert (conn);
	g_debug ("libhal_cleanup");
	DBusError error;
	dbus_error_init (&error);
	libhal_ctx_shutdown (hal_ctx, &error);
	libhal_ctx_free (hal_ctx);
	dbus_connection_disconnect (conn);
	dbus_connection_unref (conn);
}

/** Prints program usage.
 *
 */
static void print_usage (void)
{
	g_print ("\nusage : gnome-power-manager [options]\n");
	g_print (
		"\n"
		"        --system-bus            Use the gconf system bus\n"
		"        --verbose               Show extra debugging\n"
		"        --help                  Show this information and exit\n"
		"\n");
}

/** Entry point
 *
 *  @param  argc		Number of arguments given to program
 *  @param  argv		Arguments given to program
 *  @return			Return code
 */
int
main (int argc, char *argv[])
{
	gint a;
	GMainLoop *loop;
	DBusConnection *conn = NULL;
	DBusError error;
	dbus_error_init (&error);

	gconf_init (argc, argv, NULL);
	GConfClient *client = gconf_client_get_default ();
	gconf_client_add_dir (client, "/apps/gnome-power", GCONF_CLIENT_PRELOAD_NONE, NULL);
	gconf_client_notify_add (client, "/apps/gnome-power", 
		callback_gconf_key_changed, NULL, NULL, NULL);

	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);

	useSystemBus = FALSE;
	isVerbose = FALSE;
	for (a=1; a < argc; a++) {
		if (strcmp (argv[a], "--verbose") == 0)
			isVerbose = TRUE;
		else if (strcmp (argv[a], "--system-bus") == 0)
			useSystemBus = TRUE;
		else if (strcmp (argv[a], "--help") == 0) {
			print_usage ();
			return EXIT_SUCCESS;
		}
	}

	if (!isVerbose)
		g_log_set_handler (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, g_log_ignore, NULL);

	gnome_program_init ("GNOME Power Manager", VERSION, LIBGNOMEUI_MODULE, argc, argv, NULL);
	GnomeClient *master = gnome_master_client ();
	GnomeClientFlags flags = gnome_client_get_flags (master);
	if (flags & GNOME_CLIENT_IS_CONNECTED) {
		gnome_client_set_restart_style (master, GNOME_RESTART_IMMEDIATELY);
		gnome_client_flush (master);
	}
	g_signal_connect (GTK_OBJECT (master), "die", G_CALLBACK (gpm_exit), NULL);

	if (gpm_get_lock () == FALSE)
		g_error ("Already running, exiting...\n");

	g_print ("%s %s - %s\n", NICENAME, VERSION, NICEDESC);
	g_print ("Please report bugs to richard@hughsie.com\n");

	/*
	 * check for the latest set of GConf keys
	 * this will check the newest install of the GConf schema
	 */
	gchar *value = gconf_client_get_string (client, "/apps/gnome-power/scripts/shutdown", NULL);
	if (!value)
		g_warning ("The GConf schema was not installed correctly!");
	g_free (value);

	loop = g_main_loop_new (NULL, FALSE);
	conn = libhal_setup ();
	g_assert (conn);

	objectData = g_ptr_array_new ();

	coldplug_devices ();
	if (isVerbose)
		genericobject_print (objectData);

	update_has_logic (objectData, TRUE);
	update_state_logic (objectData, TRUE);

	gpn_icon_initialise (ICON_ALL);
	gpn_icon_update (ICON_ALL);
	
	dbus_connection_setup_with_g_main (conn, NULL);
	libhal_device_property_watch_all (hal_ctx, &error);

	g_main_loop_run (loop);

	/* free objectData */
	for (a=0;a<objectData->len;a++)
		g_free (g_ptr_array_index (objectData, a));
	g_ptr_array_free (objectData, TRUE);

	libhal_cleanup (conn);
	gpm_exit ();
	return 0;
}
