/***************************************************************************
 *
 * icons.c : Middle layer library for managing icons
 *
 * Copyright (C) 2005 Richard Hughes, <richard@hughsie.com>
 * Taken in part from:
 * - notibat (C) 2004 Benjamin Kahn <xkahn@zoned.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 **************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <glade/glade.h>
#include <gdk/gdk.h>
#include "eggtrayicon.h"
#include "common.h"
#include "gpm.h"
#include "icons.h"


/** private click release callback
 *
 */
static gboolean
tray_icon_release (GtkWidget *widget, GdkEventButton *event, TrayData *td)
{
	if (!td || !td->popup_menu)
		return TRUE;
	if (event->button == 3) {
		gtk_menu_popdown (GTK_MENU (td->popup_menu));
		return FALSE;
	}
	return TRUE;
}

/** private click press callback
 *
 */
static gboolean
tray_icon_press (GtkWidget *widget, GdkEventButton *event, TrayData *td)
{
	if (!td || !td->popup_menu)
		return TRUE;
	if (event->button == 3) {
		gtk_menu_popup (GTK_MENU (td->popup_menu), NULL, NULL, NULL, NULL, event->button, event->time);
		return TRUE;
	}
	return FALSE;
}

/** Creates icon in the notification area
 *
 *  @param  td			Address of the icon
 *  @param  filename		The filename of the icon
 */
void
icon_create (TrayData **td)
{
	g_assert (*td == NULL);
	g_debug ("icon_create");
	GtkWidget *evbox;

	/* create new tray object */
	*td = g_new0 (TrayData, 1);

	/* Event box */
	evbox = gtk_event_box_new ();
	(*td)->evbox = evbox;
	(*td)->tray_icon = egg_tray_icon_new (NICENAME);
	(*td)->tray_icon_tooltip = gtk_tooltips_new ();
	(*td)->popup_menu = NULL;

#if 0
	/* image */
	gchar *fullpath = g_strconcat (GPM_DATA, filename, NULL);
	(*td)->image = gtk_image_new_from_file (fullpath);
	g_free (fullpath);
#endif
	/* will produce a broken image.. */
	(*td)->image = gtk_image_new_from_file ("");
	g_signal_connect (G_OBJECT (evbox), "button_press_event", G_CALLBACK (tray_icon_press), (gpointer) *td);
	g_signal_connect (G_OBJECT (evbox), "button_release_event", G_CALLBACK (tray_icon_release), (gpointer) *td);

	gtk_container_add (GTK_CONTAINER (evbox), (*td)->image);
	gtk_container_add (GTK_CONTAINER ((*td)->tray_icon), evbox);

	gtk_widget_show_all (GTK_WIDGET ((*td)->tray_icon));
}

/** Modifies image of the notification area icon
 *
 *  @param  td			Address of the icon
 *  @param  filename		The filename of the icon
 */
void
icon_modify_image (TrayData *td, const char* filename)
{
	g_assert (td);
	g_assert (filename);
	gchar *fullpath = g_strconcat (GPM_DATA, filename, NULL);
	gtk_image_set_from_file (GTK_IMAGE (td->image), fullpath);
	g_free (fullpath);
}

/** Modifies image of the notification area icon
 *
 *  @param  td			Address of the icon
 *  @param  pixbuf		The pixbuf of the icon
 */
void
icon_modify_image_pixbuf (TrayData *td, GdkPixbuf *pixbuf)
{
	g_assert (td);
	g_assert (pixbuf);
	gtk_image_set_from_pixbuf (GTK_IMAGE (td->image), pixbuf);
}

/** Modifies tooltip of the notification area icon
 *
 *  @param  td			Address of the icon
 *  @param  tooltip		The new tooltip
 */
void
icon_modify_tooltip (TrayData *td, const char* tooltip)
{
	g_assert (td);
	g_assert (tooltip);
	g_debug ("icon_modify_tooltip (%s)", tooltip);
	gtk_tooltips_set_tip (td->tray_icon_tooltip, GTK_WIDGET (td->tray_icon), tooltip, NULL);
}

/** Frees resources and hides notification area icon
 *
 *  @param  td			Address of the icon
 *  @param  tooltip		The new tooltip
 */
void
icon_destroy (TrayData **td)
{
	g_assert (*td);
	g_debug ("icon_destroy");
	if ((*td)->popup_menu)
		g_free ((*td)->popup_menu);
	gtk_widget_hide_all (GTK_WIDGET ((*td)->tray_icon));
	g_free (*td);
	*td = NULL;
}
